// SPDX-FileCopyrightText: 2023 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

mod pb {
    tonic::include_proto!("package");
}

use {
    futures::{Stream, StreamExt as _},
    pb::service_server::{Service, ServiceServer},
    pb::Message,
    std::cell::RefCell,
    std::error::Error,
    std::pin::Pin,
    std::sync::{Arc, Mutex, Weak},
    tokio::sync::broadcast,
    tokio_stream::wrappers::BroadcastStream,
    tonic::transport::server::TcpIncoming,
    tonic::{Request, Response, Status, Streaming},
};

// How many messages to hold in memory before start discarding newcomers:
const MESSAGE_QUEUE_SIZE: usize = 20;

// gRPC server/client connect addr
const RENDEZVOUS: &'static str = "[::1]:10000";

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<(), Box<dyn Error>> {
    let url = RENDEZVOUS.parse()?;
    let incoming = TcpIncoming::new(url, true, None).expect("Cannot bind server socket");
    run_server(incoming).await?;
    Ok(())
}

fn run_server(
    incoming: TcpIncoming,
) -> impl std::future::Future<Output = Result<(), tonic::transport::Error>> {
    use tonic::transport::Server;

    let service = MessageService::default();
    let svc_adapter = ServiceServer::new(service);
    let builder = Server::builder().add_service(svc_adapter);
    builder.serve_with_incoming(incoming)
}

type ProtectedWeak<T> = Mutex<RefCell<Weak<T>>>;

#[derive(Default, Debug)]
struct MessageService {
    shared_tx: ProtectedWeak<broadcast::Sender<Message>>,
}

impl MessageService {
    // Lazily create a channel and keep a cached reference in self.shared_tx
    // for other clients to reuse. When the last client goes away, we recycle the
    // channel. This allows cleanly shutting down the server in tests too.
    fn shared_channel(
        &self,
    ) -> (
        Arc<broadcast::Sender<Message>>,
        broadcast::Receiver<Message>,
    ) {
        let tx_guard = self.shared_tx.lock().unwrap();
        let maybe_tx = tx_guard.borrow().upgrade();
        match maybe_tx {
            Some(tx) => {
                let rx = tx.subscribe();
                (tx, rx)
            }
            None => {
                let (tx, rx) = broadcast::channel(MESSAGE_QUEUE_SIZE);
                let tx = Arc::new(tx);
                *tx_guard.borrow_mut() = Arc::downgrade(&tx);
                (tx, rx)
            }
        }
    }
}

#[tonic::async_trait]
impl Service for MessageService {
    type BroadcasterStream = Pin<Box<dyn Stream<Item = Result<Message, Status>> + Send + 'static>>;

    // This is the implementation of the gRPC method from the .proto file.
    async fn broadcaster(
        &self,
        requests: Request<Streaming<Message>>,
    ) -> Result<Response<Self::BroadcasterStream>, Status> {
        let mut incoming = requests.into_inner();

        // When the connection is first established, we create a lazy stream for output messages.
        // This will be returned from the function.
        let (tx, rx) = self.shared_channel();
        let output = BroadcastStream::new(rx)
            .map(|result| result.map_err(|err| Status::data_loss(err.to_string())));

        // We also spawn a separate task to process incoming messages and
        // forward them to all clients. This will run on the executor
        // until an error is encountered; typically, until the incoming
        // stream is closed.
        tokio::spawn(async move {
            while let Some(Ok(message)) = incoming.next().await {
                let _ = tx.send(message); // Ignore err if no receivers
            }
        });

        Ok(Response::new(Box::pin(output)))
    }
}

#[cfg(test)]
mod test {
    use {
        super::pb::service_client::ServiceClient,
        super::run_server,
        super::{Message, RENDEZVOUS},
        mockall::automock,
        tonic::transport::server::TcpIncoming,
        tonic::Request,
    };

    #[automock]
    trait MessageChecker {
        fn check_contents(&self, msg: &Message);
    }

    #[tokio::test]
    async fn bidi_streaming() -> Result<(), Box<dyn std::error::Error>> {
        let url = RENDEZVOUS.parse()?;
        let incoming = TcpIncoming::new(url, true, None).expect("Cannot bind server socket");
        tokio::spawn(run_server(incoming));

        // This is the number of messages to broadcast
        const N_MESSAGES: usize = 100;
        // We use 4 test clients connecting to the server
        const CLIENT_IDS: &[&str] = &["client AAA", "client BBB", "client CCC"];

        async fn client(id: &str, mock: &MockMessageChecker) {
            let addr = format!("http://{}", RENDEZVOUS);
            let mut client = ServiceClient::connect(addr).await.unwrap();

            let client_id = String::from(id);
            let outbound = async_stream::stream! {
                for i in 1..(N_MESSAGES+1) {
                    tokio::task::yield_now().await;
                    let message = Message {
                        contents: format!("{}: {}", client_id, i),
                    };

                    yield message;
                }
            };

            let response = client.broadcaster(Request::new(outbound)).await.unwrap();
            let mut inbound = response.into_inner();
            while let Some(msg) = inbound.message().await.unwrap() {
                println!("{} received {:?}", id, msg);
                mock.check_contents(&msg);
            }
        }

        // We create a mock to check that we receive exactly the
        // messages we want. We also check for saturation.
        let mut mock = MockMessageChecker::new();
        for client_id in CLIENT_IDS.iter() {
            for i in 1..(N_MESSAGES + 1) {
                let expected = format!("{}: {}", client_id, i);
                mock.expect_check_contents()
                    .times(CLIENT_IDS.len())
                    .withf(move |msg| msg.contents == expected)
                    .return_const(());
            }
        }

        // Time to create the clients, and await for their completion
        let clients = CLIENT_IDS
            .into_iter()
            .map(|&id| client(id, &mock))
            .collect::<Vec<_>>();
        let _ = futures::future::join_all(clients).await;

        println!("Exiting.");
        Ok(())
    }
}
